import argparse
import logging
import os
from typing import Optional

from spotipy import Spotify
from spotipy.util import prompt_for_user_token

from .search import create_spotify_playlist, update_spotify_playlist


def _m3u_file(path: str) -> Optional[str]:
    if not isinstance(path, str):
        raise argparse.ArgumentTypeError('Path must be a string. Encountered type={}'.format(str(type(path))))

    splits = path.rsplit('.', 1)
    if len(splits) == 1:
        raise argparse.ArgumentTypeError('Could not determine file extension')

    filename, extension = splits[0], splits[1]
    if len(filename) == 0:
        raise argparse.ArgumentTypeError('Filename without extension cannot be empty')

    if extension in {'m3u', 'm3u8'}:
        return path

    raise argparse.ArgumentTypeError('Only m3u files are supported. Encountered={}'.format(extension))


def _parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description='Create/update a Spotify playlist from a local m3u playlist')
    parser.add_argument('spotify_username',
                        help='Spotify username where playlist should be updated. '
                             'Your email address should work just fine, or could find your user id '
                             'through e.g. the developer console', type=str)
    parser.add_argument('playlist_path', help='full path to the playlist', type=_m3u_file)
    parser.add_argument('--verbose', help='print all the steps when searching for songs', action='store_true')
    parser.add_argument('--public', help='playlist is public, otherwise private', action='store_true')
    parser.add_argument('--convert', help='convert from locale default to utf-8', action='store_true')
    parser.add_argument('--playlist-id', help='do not create a new playlist, '
                                              'instead update the existing playlist with this id', type=str)
    parsed_args = parser.parse_args()

    return parsed_args


def _parse_path(path: str) -> str:
    if path.startswith('/'):
        return path

    return os.path.join(os.getcwd(), *path.split(os.sep))


def _extract_playlist_name(playlist_path: str) -> str:
    _, filename = os.path.split(playlist_path)
    playlist_name = str(filename.split('.')[0])

    return playlist_name


def main() -> None:
    """ Main entry point to the script """
    args = _parse_args()

    logger = logging.getLogger()
    logger_level = logging.INFO if args.verbose else logging.WARNING
    logger.setLevel(logger_level)

    username = args.spotify_username
    scope = 'playlist-modify-public' if args.public else 'playlist-modify-private'
    token = prompt_for_user_token(username, scope)
    playlist_path = _parse_path(args.playlist_path)

    if token is None:
        logging.error('Could not acquire token!')

    spot = Spotify(auth=token)

    if args.playlist_id is None:
        playlist_name = _extract_playlist_name(playlist_path)
        playlist_id = create_spotify_playlist(spot, playlist_name)
        logging.info('Created playlist with name={} at id={}'.format(playlist_name, playlist_id))
    else:
        playlist_id = args.playlist_id
        logging.info('Updating existing playlist with id={}'.format(playlist_id))

    update_spotify_playlist(spot, playlist_path, playlist_id, args.convert)



import logging
import re
from typing import Tuple


MIN_LENGTH_NAME = 3


def clean_title(title: str) -> str:
    """ Clean a song title:

    - remove brackets and text within ()[] brackets, e.g. [acoustic]
    - remove text (and including) feat. or featuring or ft.
    - strip the remaining whitespaces

    :param title: song title to clean
    :type title: str
    :return: cleaned song title
    :rtype: str
    """
    cleaned_title = re.sub(r'\([^)]*\)|\[[^)]*\]', '', title)
    cleaned_title = re.sub(r'(\sfeat\..*)|(\sfeaturing.*)|(\sft\..*)', '', cleaned_title)
    cleaned_title = cleaned_title.strip()

    return cleaned_title


def clean_name(name: str) -> str:
    """ Clean either artist or title:

    - removes these characters: .{}
    - removes extra spaces
    - strip and lowercase

    :param name: artist or song title to clean
    :type name: str
    :return: cleaned name
    :rtype: str
    """
    cleaned_name = re.sub(r'[.{\}]', '', name)
    cleaned_name = re.sub(r'\s+', ' ', cleaned_name)
    cleaned_name = cleaned_name.strip()
    cleaned_name = cleaned_name.lower()

    if len(cleaned_name) == 0:
        logging.warning('Encountered empty string after cleaning. Original string={}'.format(name))

    return cleaned_name


def process_song_name(song_name: str) -> Tuple[str, str]:
    """ Splits m3u line of artist - title and cleans using clean_name.
    Note that the Extended M3U8 format requires this format, i.e. artist - title based on file tags.

    !Since '-' delimits the artist from the song, this character should not be seen inside the artist or title!

    :param song_name:
    :type song_name: str
    :return: tuple of artist and title
    :rtype: (str, str)
    :raises ProcessingException: if a song does not obey the Extended M3U8 formatting of "artist - title"
    """
    song_split = song_name.split('-')

    if len(song_split) == 1:
        raise ProcessingException('Could not split song into artist and title! song={}'.format(song_name))

    if len(song_split) != 2:
        logging.warning('Encountered more than 2 chunks when splitting song into artist and title. '
                        'song={}'.format(song_name))

    artist = clean_name(song_split[0])
    title = clean_name(song_split[1])

    return artist, title


class ProcessingException(Exception):
    """ Custom exception for errors in processing the data """



import logging
from typing import Generator

from ..processing import MIN_LENGTH_NAME


SEP_SEMICOLON = ';'
# added spaces to avoid and being inside a word, e.g. andrew
SEP_AND = ' and '
SEP_AND_SYMBOL = ' & '
TOKEN_THE = 'the '  # nosec


class Artist:
    """ Base class for an Artist """
    def __init__(self, artist: str) -> None:
        self.artist = artist

    def makes_sense(self) -> bool:  # pylint: disable=no-self-use
        """ Checks whether it makes sense to compile a query given this artist

        :return: True if a query would make sense, False otherwise
        :rtype: bool
        """
        return True

    def process(self) -> Generator[str, None, None]:
        """ Processes the artist according to the implemented rules.
        Note it might result in multiple titles to try.

        :return: processed title
        :rtype: str
        """
        yield self.artist


class ArtistBeginsWithThe(Artist):
    """ Artist class which will clean "the" from the artist name if it begins as such,
    e.g. The Gipsy Kings -> Gipsy Kings
    """
    def makes_sense(self) -> bool:
        return self.artist.lower().startswith(TOKEN_THE)

    def process(self) -> Generator[str, None, None]:
        artist = self.artist[len(TOKEN_THE):]

        if len(artist) < MIN_LENGTH_NAME:
            logging.warning('Encountered artist after removing "the" with '
                            'too short name={}'.format(artist))

        yield artist


class MultipleArtist(Artist):
    """ Base class for an Artist actually containing multiple artists separated by some separator """
    def __init__(self, artist: str, sep: str) -> None:
        self.sep = sep
        super().__init__(artist)

    def makes_sense(self) -> bool:
        return self.sep in self.artist.lower()


class MultipleTogetherArtist(MultipleArtist):
    """ MultipleArtist class which will try converting the separator to another separator,
    e.g. Sting & The Police -> Sting and The Police
    """
    def __init__(self, artist: str, sep: str, to_sep: str) -> None:
        self.to_sep = to_sep
        super().__init__(artist, sep)

    def process(self) -> Generator[str, None, None]:
        artist = self.artist.replace(self.sep, self.to_sep, 1)
        if self.sep in artist:
            logging.warning('Encountered more than 2 occurrences of sep={} when detecting artist '
                            'as multiple artists. artist={}'.format(self.sep, self.artist))

        yield artist


class MultipleTogetherAndArtist(MultipleTogetherArtist):
    """ MultipleTogetherArtist class where separator is "and" """
    def __init__(self, artist: str) -> None:
        super().__init__(artist, SEP_AND, SEP_AND_SYMBOL)


class MultipleTogetherAndSymbolArtist(MultipleTogetherArtist):
    """ MultipleTogetherArtist class where separator is "&" """
    def __init__(self, artist: str) -> None:
        super().__init__(artist, SEP_AND_SYMBOL, SEP_AND)


class MultipleSplitArtist(MultipleArtist):
    """ MultipleArtist class which will try splitting the artist string into its actual artists,
    e.g. Sting & The Police -> Sting, The Police
    """
    def process(self) -> Generator[str, None, None]:
        artists = self.artist.split(self.sep)
        for artist in artists:
            artist = artist.strip()

            if len(artist) < MIN_LENGTH_NAME:
                logging.warning('Encountered artist after splitting on={} '
                                'with too short name={}'.format(self.sep, artist))

            yield artist


class MultipleSplitSemicolonArtist(MultipleSplitArtist):
    """ MultipleSplitArtist class where separator is ";" """
    def __init__(self, artist: str) -> None:
        super().__init__(artist, SEP_SEMICOLON)


class MultipleSplitAndArtist(MultipleSplitArtist):
    """ MultipleSplitArtist class where separator is "and" """
    def __init__(self, artist: str) -> None:
        super().__init__(artist, SEP_AND)


class MultipleSplitAndSymbolArtist(MultipleSplitArtist):
    """ MultipleSplitArtist class where separator is "&" """
    def __init__(self, artist: str) -> None:
        super().__init__(artist, SEP_AND_SYMBOL)



import abc
from typing import Generator

from .artists import Artist
from .titles import Title

SEP_SEMICOLON = ';'
# added spaces to avoid and being inside a word, e.g. andrew
SEP_AND = ' and '
SEP_AND_SYMBOL = ' & '
TOKEN_THE = 'the '  # nosec


class Query(abc.ABC):
    """ Abstract class for a type of query """
    def __init__(self, artist: Artist, title: Title) -> None:
        self.artist = artist
        self.title = title

    def makes_sense(self) -> bool:
        """ Checks whether it makes sense to compile a query using its artist and title

        :return: true if it makes sense, false otherwise
        :rtype: bool
        """
        return self.artist.makes_sense() and self.title.makes_sense()

    @abc.abstractmethod
    def compile(self) -> Generator[str, None, None]:
        """ Compiles a valid Spotify search query parameter.

        :return: a Spotify search query parameter
        :rtype: str
        """
        raise NotImplementedError

    @staticmethod
    def _surround_with_quotes(string: str) -> str:
        return '\"' + string + '\"'

    @staticmethod
    def _prepare_query_component(string: str, query_type: str) -> str:
        # need to escape the quotes else requests will url-encode them
        return query_type + ':' + Query._surround_with_quotes(string)

    def _prepare_artist_component(self) -> Generator[str, None, None]:
        for artist in self.artist.process():
            yield Query._prepare_query_component(artist, 'artist')

    def _prepare_title_component(self) -> str:
        return Query._prepare_query_component(self.title.process(), 'track')


class QueryArtistTitle(Query):
    """ Query which specifically looks for a given artist and a given title """
    def compile(self) -> Generator[str, None, None]:
        title = self._prepare_title_component()
        for artist in self._prepare_artist_component():
            yield ' '.join([artist, 'AND', title])


class QueryWildcard(Query):
    """ Query which lets Spotify returns its best result based on the artist and title """
    def compile(self) -> Generator[str, None, None]:
        title = self.title.process()
        for artist in self.artist.process():
            yield Query._surround_with_quotes(artist + ' ' + title)



from setuptools import setup


def readme():
    with open('README.md') as f:
        return f.read()


setup(
    name='tospotify',
    version='0.3',
    description='Create/update a Spotify playlist from a local m3u playlist',
    url='https://github.com/radujica/tospotify',
    author='Radu Jica',
    author_email='radu.jica+code@gmail.com',
    license='GPL-3.0',
    long_description=readme(),
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Multimedia :: Sound/Audio',
        'Operating System :: OS Independent'
    ],
    packages=['tospotify', 'tospotify.types'],
    include_package_data=True,
    zip_safe=True,
    install_requires=['spotipy', 'm3u8>=0.7.1'],
    entry_points={
        'console_scripts': ['tospotify=tospotify.run:main'],
    }
)



import logging

from ..processing import clean_title, MIN_LENGTH_NAME


class Title:
    """ Base class for a song title """
    def __init__(self, title: str) -> None:
        self.title = title

    def makes_sense(self) -> bool:  # pylint: disable=no-self-use
        """ Checks whether it makes sense to compile a query given this song title

        :return: True if a query would make sense, False otherwise
        :rtype: bool
        """
        return True

    def process(self) -> str:
        """ Processes the title according to the implemented rules.

        :return: processed title
        :rtype: str
        """
        return self.title


class CleanedTitle(Title):
    """ Title class which will clean the song title if it makes sense,
    e.g. Every breath feat. Sting -> Every breath
    """
    def makes_sense(self) -> bool:
        return clean_title(self.title) != self.title

    def process(self) -> str:
        title = clean_title(self.title)

        if len(title) < MIN_LENGTH_NAME:
            logging.warning('Encountered title after cleaning with too short name={}'.format(title))

        return title



from typing import Tuple, Type, Generator

from .types.artists import (
    Artist,
    ArtistBeginsWithThe,
    MultipleTogetherAndArtist,
    MultipleTogetherAndSymbolArtist,
    MultipleSplitSemicolonArtist,
    MultipleSplitAndArtist,
    MultipleSplitAndSymbolArtist
)
from .types.queries import Query, QueryArtistTitle, QueryWildcard
from .types.titles import Title, CleanedTitle

ARTIST_POOL = (Artist, ArtistBeginsWithThe, MultipleTogetherAndArtist, MultipleTogetherAndSymbolArtist,
               MultipleSplitSemicolonArtist, MultipleSplitAndArtist, MultipleSplitAndSymbolArtist)
TITLE_POOL = (Title, CleanedTitle)
QUERY_POOL = (QueryArtistTitle, QueryWildcard)


class QueryGenerator:
    """ Generator of relevant queries for a artist-title pair """
    def __init__(
            self,
            artist: str,
            title: str,
            queries: Tuple[Type[Query]] = QUERY_POOL,
            artists: Tuple[Type[Artist]] = ARTIST_POOL,
            titles: Tuple[Type[Title]] = TITLE_POOL
    ) -> None:
        self.artist = artist
        self.title = title
        self.queries = queries
        self.artists = artists
        self.titles = titles

    def generate(self) -> Generator[str, None, None]:
        """ Return a query at a time if it would make sense to execute

        :return: query as a string
        :rtype: str
        """
        for q_class in self.queries:
            for a_class in self.artists:
                artist = a_class(self.artist)
                for t_class in self.titles:
                    title = t_class(self.title)
                    query = q_class(artist, title)
                    if query.makes_sense():
                        for query2 in query.compile():
                            yield query2



import logging

import m3u8


def convert_utf8(playlist_path: str) -> str:
    """ Convert file to utf-8 after parsing with locale.getpreferredencoding

    :param playlist_path: absolute path of the playlist
    :return: Path of converted file
    """
    logging.warning('Converting file to utf-8')
    path_without_extension = playlist_path.rsplit('.', 1)[0]
    output_path = path_without_extension + '_utf8.m3u8'

    # here it uses the locale.getpreferredencoding, which could be cp1252 for Windows
    with open(playlist_path, mode='r') as input_:
        with open(output_path, encoding='utf-8', mode='w') as output_:
            for line in input_.readlines():
                output_.write(line.encode('utf-8').decode('utf-8'))

    return output_path


def parse_songs(playlist_path: str) -> m3u8.SegmentList:
    """ Parse and return the songs found in the file

    :param playlist_path: absolute path of the playlist
    :type playlist_path: str
    :return:
    """
    # m3u8 uses open(..., encoding='utf-8') which will through exception when the file cannot be parsed as utf-8
    playlist = m3u8.load(playlist_path)
    segments = playlist.segments

    if len(segments) <= 0:
        logging.error('Could not find any songs in the file!')

    return segments



import logging
from typing import List, Optional

import m3u8
from spotipy import Spotify

from .generator import QueryGenerator
from .parser import parse_songs, convert_utf8
from .processing import process_song_name, MIN_LENGTH_NAME


def get_user_id(spot: Spotify) -> str:
    """ Return user id of this Spotify connection

    :param spot: spotipy Spotify object to use its REST API
    :type spot: Spotify
    :return: str
    :rtype: str
    """
    return spot.current_user()['id']


def create_spotify_playlist(spot: Spotify, playlist_name: str, public: bool = False) -> str:
    """ Create a playlist on Spotify

    :param spot: spotipy Spotify object to use its REST API
    :type spot: Spotify
    :param playlist_name: name of future the playlist
    :type playlist_name: str
    :param public: public or private playlist; note this assumes sp was created using the appropriate scope
    :type public: bool
    :return: the id of the just-created playlist
    :rtype: str
    """
    user_id = get_user_id(spot)
    res = spot.user_playlist_create(user_id, playlist_name, public=public)
    playlist_id = res['id']

    return playlist_id


def _run_query(spot: Spotify, query: str, market: str = None, iteration: int = 0) -> Optional[str]:
    response = spot.search(query, limit=1, type='track', market=market)
    results = response['tracks']['items']
    prepend_pretty_print = '-' * iteration
    uri = None
    if len(results) > 0:
        uri = results[0]['uri']
        logging.info('{}Found track with query={} as uri={}'.format(prepend_pretty_print, query, uri))
    else:
        logging.info('{}Could not find track with query={}'.format(prepend_pretty_print, query))

    return uri


# the code is optimized to limit the number of queries to Spotify as much as possible
def _find_track(spot: Spotify, song: m3u8.Segment, market: str = None) -> Optional[str]:
    artist, title = process_song_name(song.title())

    if len(artist) < MIN_LENGTH_NAME:
        logging.warning('Encountered artist after processing with short name={}'.format(artist))
    if len(title) < MIN_LENGTH_NAME:
        logging.warning('Encountered title after processing with short name={}'.format(title))

    query_generator = QueryGenerator(artist, title)
    iteration = 0
    for query in query_generator.generate():
        uri = _run_query(spot, query, market=market, iteration=iteration)

        if uri is not None:
            return uri

        iteration += 1

    return None


def add_tracks(spot: Spotify, playlist_id: str, tracks: List[str]) -> None:
    """ Add tracks to a Spotify playlist

    :param spot: spotipy Spotify object to use its REST API
    :type spot: Spotify
    :param playlist_id: id of the playlist
    :type playlist_id: str
    :param tracks: list of Spotify track uri's to add
    :type tracks: [str]
    :return:
    """
    user_id = get_user_id(spot)
    max_tracks_per_request = 100
    for i in range(len(tracks) // max_tracks_per_request + 1):
        tracks_subset = tracks[i * max_tracks_per_request:(i + 1) * max_tracks_per_request]
        spot.user_playlist_add_tracks(user_id, playlist_id, tracks_subset)


def update_spotify_playlist(
        spot: Spotify,
        playlist_path: str,
        playlist_id: str,
        to_convert: bool,
        market: str = None
) -> None:
    """ Read songs from an m3u file and update a Spotify playlist with these songs

    :param spot: spotipy Spotify object to use its REST API
    :type spot: Spotify
    :param playlist_path: absolute path of the playlist
    :type playlist_path: str
    :param playlist_id: Spotify id for the playlist to update
    :type playlist_id: str
    :param to_convert: whether to convert file to utf-8
    :type to_convert: bool
    :param market: market to look for songs in; see Spotipy docs
    :type market: str
    :return:
    """
    if to_convert:
        playlist_path = convert_utf8(playlist_path)

    songs = parse_songs(playlist_path)

    if len(songs) > 0:
        tracks = []
        for song in songs:
            track_uri = _find_track(spot, song.title, market)
            if track_uri:
                tracks.append(track_uri)
            else:
                logging.warning('Could not find any track for song with artist - title={}'.format(song.title))
        if len(tracks) == 0:
            logging.error('Could not find any tracks on Spotify!')
        else:
            add_tracks(spot, playlist_id, tracks)
    else:
        logging.info('Did not try searching on Spotify since no songs were found in the file.')
